"""Description of Pd2d class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
from warnings import warn
import numpy
from typing import List, NoReturn

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_cos_ang
from cryspy.A_functions_base.function_1_matrices import calc_mRmCmRT

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN

from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_phase import PhaseL
from cryspy.C_item_loop_classes.cl_1_refln import ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs
from cryspy.C_item_loop_classes.cl_1_pd2d_background import \
    Pd2dBackground
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_reflex_asymmetry import\
    Pd2dInstrReflexAsymmetry
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_resolution import \
    Pd2dInstrResolution
from cryspy.C_item_loop_classes.cl_1_pd2d_meas import Pd2dMeas
from cryspy.C_item_loop_classes.cl_1_pd2d_proc import Pd2dProc
from cryspy.C_item_loop_classes.cl_1_pd2d_peak import Pd2dPeakL
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2
from cryspy.C_item_loop_classes.cl_1_range import Range
from cryspy.C_item_loop_classes.cl_1_texture import Texture
from cryspy.C_item_loop_classes.cl_1_exclude import ExcludeL

from cryspy.E_data_classes.cl_1_crystal import Crystal


class Pd2d(DataN):
    """
    Pd2d class.

    Data items in the DIFFRN category record details about
    2d powder diffraction measurements.

    Methods
    -------
        - calc_iint_u_d_flip_ratio
        - calc_fr
        - calc_fm_perp_loc
        - calc_chi_sq
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - estimate_FM


    Attributes
    ----------
        - setup (mandatory)
        - pd2d_instr_resolution (mandatory)
        - phase (mandatory)
        - pd2d_background (mandatory)
        - pd_meas (mandatory)
        - diffrn_radiation
        - chi2
        - range
        - extinction
        - pd2d_instr_reflex_asymmetry
        - texture
        - exclude
        - pd2d_proc
        - pd2d_peak
        - refine_ls
        - refln_#phase_name
        - refln_susceptibility_#phase_name
    """

    CLASSES_MANDATORY = (Pd2dInstrResolution, PhaseL, DiffrnRadiation,
                         Setup, Range, Chi2, Pd2dBackground, Pd2dMeas)
    CLASSES_OPTIONAL = (Extinction, ExcludeL, Pd2dInstrReflexAsymmetry,
                        Texture, RefineLs, ReflnL, ReflnSusceptibilityL,
                        Pd2dPeakL, Pd2dProc)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "pd2d"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, data_name=None, **kwargs) -> NoReturn:
        super(Pd2d, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["data_name"] = data_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def calc_profile(self, tth, phi, l_crystal: List[Crystal], l_peak_in=None,
                     l_refln_in=None, l_refln_susceptibility_in=None,
                     l_dd_in=None, flag_internal: bool = True):
        """Calculate intensity for the given diffraction angle."""
        proc = Pd2dProc()  # it is output
        proc.ttheta = tth
        proc.phi = phi

        background = self.pd2d_background
        int_bkgd = background.interpolate_by_points(tth, phi)
        proc.intensity_bkg_calc = int_bkgd

        tth_rad = tth*numpy.pi/180.
        phi_rad = phi*numpy.pi/180.
        cos_theta_1d = numpy.cos(0.5*tth_rad)
        sin_phi_1d = numpy.sin(phi_rad)

        setup = self.setup
        wavelength = setup.wavelength
        diffrn_radiation = self.diffrn_radiation
        if setup.offset_phi is None:
            setup.offset_phi = 0.

        phi_0 = setup.offset_phi
        phi_rad = (phi-phi_0)*numpy.pi/180.
        sin_phi_1d = numpy.sin(phi_rad)

        p_u = float(diffrn_radiation.polarization)
        p_d = (2.*float(diffrn_radiation.efficiency)-1.)*p_u

        tth_min = tth.min()
        tth_max = tth.max()+3.
        if tth_max > 180.:
            tth_max = 180.

        sthovl_min = numpy.sin(0.5*tth_min*numpy.pi/180.)/wavelength
        sthovl_max = numpy.sin(0.5*tth_max*numpy.pi/180.)/wavelength

        res_u_2d = numpy.zeros((tth.shape[0], phi.shape[0]), dtype=float)
        res_d_2d = numpy.zeros((tth.shape[0], phi.shape[0]), dtype=float)

        try:
            texture = self.texture
            h_ax, k_ax = float(texture.h_ax), float(texture.k_ax)
            l_ax = float(texture.l_ax)
            g_1, g_2 = float(texture.g_1), float(texture.g_2)
        except AttributeError:
            texture = None

        phase = self.phase
        _h = len(phase.label)
        if l_peak_in is None:
            l_peak_in = len(phase.label) * [None]
        else:
            if _h != len(l_peak_in):
                l_peak_in = len(phase.label) * [None]
        if l_refln_in is None:
            l_refln_in = len(phase.label) * [None]
        else:
            if _h != len(l_refln_in):
                l_refln_in = len(phase.label) * [None]
        if l_refln_susceptibility_in is None:
            l_refln_susceptibility_in = len(phase.label) * [None]
        else:
            if _h != len(l_refln_susceptibility_in):
                l_refln_susceptibility_in = len(phase.label) * [None]
        if l_dd_in is None:
            l_dd_in = len(phase.label) * [None]
        else:
            if _h != len(l_dd_in):
                l_dd_in = len(phase.label) * [None]

        l_peak, l_dd_out = [], []
        l_refln, l_refln_s = [], []
        for item_phase, peak_in, refln_in, refln_susceptibility_in, dd_in in \
                zip(phase.items, l_peak_in, l_refln_in,
                    l_refln_susceptibility_in, l_dd_in):
            phase_label = item_phase.label
            phase_scale = item_phase.scale
            try:
                phase_igsize = item_phase.igsize
                if phase_igsize is None: phase_igsize = 0. # temporary solution
            except AttributeError:
                phase_igsize = 0.
            try:
                phase_u = item_phase.u
                if phase_u is None: phase_u = 0. # temporary solution
            except AttributeError:
                phase_u = 0.
            try:
                phase_v = item_phase.v
                if phase_v is None: phase_v = 0. # temporary solution
            except AttributeError:
                phase_v = 0.
            try:
                phase_w = item_phase.w
                if phase_w is None: phase_w = 0. # temporary solution
            except AttributeError:
                phase_w = 0.
            try:
                phase_x = item_phase.x
                if phase_x is None: phase_x = 0. # temporary solution
            except AttributeError:
                phase_x = 0.
            try:
                phase_y = item_phase.y
                if phase_y is None: phase_y = 0. # temporary solution
            except AttributeError:
                phase_y = 0.

            dd_out = {}
            for i_crystal, crystal in enumerate(l_crystal):
                if crystal.data_name.lower() == phase_label.lower():
                    ind_cry = i_crystal
                    break
            if ind_cry is None:
                warn(f"Crystal with name '{phase_label:}' is not found.",
                     UserWarning)
                return

            crystal = l_crystal[ind_cry]

            cell = crystal.cell
            space_group = crystal.space_group

            if peak_in is not None:
                index_h = peak_in.numpy_index_h
                index_k = peak_in.numpy_index_k
                index_l = peak_in.numpy_index_l
                mult = peak_in.numpy_index_mult
            else:
                if texture is None:
                    index_h, index_k, index_l, mult = cell.calc_hkl(
                        space_group, sthovl_min, sthovl_max)
                else:
                    index_h, index_k, index_l, mult = cell.calc_hkl_in_range(
                        sthovl_min, sthovl_max)

            peak = Pd2dPeakL(loop_name=phase_label)
            peak.numpy_index_h = index_h
            peak.numpy_index_k = index_k
            peak.numpy_index_l = index_l
            peak.numpy_index_mult = mult
            peak.numpy_to_items()

            cond_1 = len(crystal.get_variable_names()) == 0
            cond_2 = (peak_in is not None) & (refln_in is not None)

            if (cond_1 & cond_2):
                f_nucl_sq = peak_in.numpy_f_nucl_sq
                f_m_p_sin_sq = peak_in.numpy_f_m_p_sin_sq
                f_m_p_cos_sq = peak_in.numpy_f_m_p_cos_sq
                cross_sin = peak_in.numpy_cross_sin
                refln = refln_in
                refln_s = refln_susceptibility_in
            else:
                f_nucl_sq, f_m_p_sin_sq, f_m_p_cos_sq, cross_sin, refln, \
                    refln_s = self.calc_for_iint(index_h, index_k, index_l,
                                                 crystal)
                refln.loop_name = phase_label
                refln_s.loop_name = phase_label
            l_refln.append(refln)
            l_refln_s.append(refln_s)

            peak.numpy_f_nucl_sq = f_nucl_sq
            peak.numpy_f_m_p_sin_sq = f_m_p_sin_sq
            peak.numpy_f_m_p_cos_sq = f_m_p_cos_sq
            peak.numpy_cross_sin = cross_sin

            cond_1 = dd_in is not None
            cond_2 = ((len(crystal.get_variable_names()) == 0) &
                      (not(diffrn_radiation.polarization_refinement)) &
                      (not(diffrn_radiation.efficiency_refinement)))
            if cond_1 & cond_2:
                iint_u_3d, iint_d_3d = dd_in["iint_u_3d"], dd_in["iint_d_3d"]
                cos_theta_3d = dd_in["cos_theta_3d"]
                sin_phi_3d = dd_in["sin_phi_3d"]
            else:
                cos_theta_3d, sin_phi_3d, mult_f_n_3d = numpy.meshgrid(
                    cos_theta_1d, sin_phi_1d, mult*f_nucl_sq, indexing="ij")

                mult_f_m_c_3d = numpy.meshgrid(
                    tth_rad, phi_rad, mult*f_m_p_cos_sq, indexing="ij")[2]

                hh_u_s_3d = numpy.meshgrid(
                    tth_rad, phi_rad, mult*(f_m_p_sin_sq+p_u*cross_sin),
                    indexing="ij")[2]
                hh_d_s_3d = numpy.meshgrid(
                    tth_rad, phi_rad, mult*(f_m_p_sin_sq-p_d*cross_sin),
                    indexing="ij")[2]

                c_a_sq_3d = (cos_theta_3d * sin_phi_3d)**2
                s_a_sq_3d = 1.-c_a_sq_3d
                iint_u_3d = (mult_f_n_3d + hh_u_s_3d*s_a_sq_3d +
                             mult_f_m_c_3d*c_a_sq_3d)
                iint_d_3d = (mult_f_n_3d + hh_d_s_3d*s_a_sq_3d +
                             mult_f_m_c_3d*c_a_sq_3d)
            dd_out["cos_theta_3d"] = cos_theta_3d
            dd_out["sin_phi_3d"] = sin_phi_3d
            dd_out["iint_u_3d"] = iint_u_3d
            dd_out["iint_d_3d"] = iint_d_3d

            sthovl_hkl = cell.calc_sthovl(index_h, index_k, index_l)
            tth_hkl_rad = numpy.where(sthovl_hkl*wavelength < 1.,
                                      2.*numpy.arcsin(sthovl_hkl*wavelength),
                                      numpy.pi)
            tth_hkl = tth_hkl_rad*180./numpy.pi

            cond_1 = dd_in is not None
            flag_phase_uvwxy_ref = (
                item_phase.igsize_refinement | item_phase.u_refinement |
                item_phase.v_refinement | item_phase.w_refinement |
                item_phase.x_refinement | item_phase.y_refinement)
            cond_2 = ((not(flag_phase_uvwxy_ref)) &
                      (not(self.pd2d_instr_resolution.is_variables())) &
                      (not(setup.is_variables())))
            if cond_1 & cond_2:
                profile_3d, tth_zs = dd_in["profile_3d"], dd_in["tth_zs"]
                h_pv = dd_in["h_pv"]
            else:
                profile_3d, tth_zs, h_pv = self.calc_shape_profile(
                    tth, phi, tth_hkl, phase_igsize=phase_igsize,
                    phase_u=phase_u, phase_v=phase_v, phase_w=phase_w,
                    phase_x=phase_x, phase_y=phase_y)
            dd_out["profile_3d"] = profile_3d
            dd_out["tth_zs"] = tth_zs
            dd_out["h_pv"] = h_pv

            peak.numpy_ttheta = tth_hkl + setup.offset_ttheta
            peak.width_ttheta = h_pv
            peak.numpy_to_items()

            # texture
            if texture is not None:
                cond_1 = dd_in is not None
                cond_2 = ((not(setup.offset_phi_refinement)))
                if cond_1 & cond_2:
                    cos_alpha_ang_3d = dd_in["cos_alpha_ang_3d"]
                    sin_alpha_ang_3d = dd_in["sin_alpha_ang_3d"]
                else:
                    cos_alpha_ang_3d = cos_theta_3d * sin_phi_3d
                    sin_alpha_ang_3d = numpy.sqrt(1.-cos_alpha_ang_3d**2)
                dd_out["cos_alpha_ang_3d"] = cos_alpha_ang_3d
                dd_out["sin_alpha_ang_3d"] = sin_alpha_ang_3d
                cond_2 = (cond_2 & (not(texture.is_variables())) &
                          (not(cell.is_variables())))
                if cond_1 & cond_2:
                    texture_3d = dd_in["texture_3d"]
                else:
                    cos_alpha_ax = calc_cos_ang(cell, h_ax, k_ax, l_ax,
                                                index_h, index_k, index_l)
                    c_help = 1.-cos_alpha_ax**2
                    c_help[c_help < 0.] = 0.
                    sin_alpha_ax = numpy.sqrt(c_help)
                    cos_alpha_ax_3d = numpy.meshgrid(tth, phi, cos_alpha_ax,
                                                     indexing="ij")[2]
                    sin_alpha_ax_3d = numpy.meshgrid(tth, phi, sin_alpha_ax,
                                                     indexing="ij")[2]
                    cos_alpha_3d = cos_alpha_ax_3d*cos_alpha_ang_3d + \
                        sin_alpha_ax_3d*sin_alpha_ang_3d
                    texture_3d = g_2 + (1.-g_2) * \
                        (1./g_1 + (g_1**2-1./g_1)*cos_alpha_3d**2)**(-1.5)
                dd_out["texture_3d"] = texture_3d

                profile_3d = profile_3d*texture_3d

            res_u_3d = profile_3d*iint_u_3d
            res_d_3d = profile_3d*iint_d_3d

            # 0.5 to have the same meaning for scale factor as in FullProf
            res_u_2d += 0.5*phase_scale*res_u_3d.sum(axis=2)
            res_d_2d += 0.5*phase_scale*res_d_3d.sum(axis=2)
            l_peak.append(peak)
            l_dd_out.append(dd_out)

        proc.ttheta_corrected = tth_zs
        proc.intensity_up_net = res_u_2d
        proc.intensity_down_net = res_d_2d
        proc.intensity_up_total = res_u_2d+int_bkgd
        proc.intensity_down_total = res_d_2d+int_bkgd
        if flag_internal:
            l_calc_objs = l_refln + l_refln_s + l_peak
            proc.form_ttheta_phi_intensity_up_net()
            proc.form_ttheta_phi_intensity_down_net()
            proc.form_ttheta_phi_intensity_up_total()
            proc.form_ttheta_phi_intensity_down_total()
            l_calc_objs.append(proc)
            self.add_items(l_calc_objs)

        return proc, l_peak, l_refln, l_dd_out

    def calc_chi_sq(self, l_crystal, flag_internal: bool = True):
        """
        Calculate chi square.

        Arguments
        ---------
            - l_crystal: a list of Crystal objects of cryspy library
            - flag_internal: a flag to calculate internal objects
              (default is True)

        Output arguments
        ----------------
            - chi_sq_val: chi square of flip ratio
              (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
            - n: number of measured reflections
        """
        meas = self.pd2d_meas

        tth = meas.ttheta
        phi = meas.phi
        int_u_exp = meas.intensity_up
        sint_u_exp = meas.intensity_up_sigma
        int_d_exp = meas.intensity_down
        sint_d_exp = meas.intensity_down_sigma

        l_peak_in, l_refln_in = [], []
        l_refln_susceptibility_in = []
        l_dd_in = []
        flag_1 = not(flag_internal)

        try:
            dd = self.dd
            flag_2 = True
        except AttributeError:
            flag_2 = False
        if (flag_1 & flag_2):
            for phase_item in self.phase.items:
                crystal = None
                for cryst in l_crystal:
                    if cryst.data_name.lower() == phase_item.label.lower():
                        crystal = cryst
                        break
                attr_peak = f"pd2d_peak_{crystal.data_name:}"
                attr_refln = f"refln_{crystal.data_name:}"
                attr_refln_s = f"refln_susceptibility_{crystal.data_name:}"
                l_peak_in.append(getattr(self, attr_peak))
                l_refln_in.append(getattr(self, attr_refln))
                l_refln_susceptibility_in.append(getattr(self, attr_refln_s))

        cond_tth_in = numpy.ones(tth.size, dtype=bool)
        cond_phi_in = numpy.ones(phi.size, dtype=bool)
        try:
            range_ = self.range
            cond_tth_in = numpy.logical_and(cond_tth_in, tth >=
                                            range_.ttheta_min)
            cond_tth_in = numpy.logical_and(cond_tth_in, tth <=
                                            range_.ttheta_max)

            cond_phi_in = numpy.logical_and(cond_phi_in, phi >= range_.phi_min)
            cond_phi_in = numpy.logical_and(cond_phi_in, phi <= range_.phi_max)

        except AttributeError:
            pass
        # cond_1_in, cond_2_in = numpy.meshgrid(cond_tth_in, cond_phi_in,
        #                                       indexing="ij")
        # cond_in = numpy.logical_and(cond_1_in, cond_2_in)
        tth_in = tth[cond_tth_in]
        phi_in = phi[cond_phi_in]
        int_u_exp_in = int_u_exp[cond_tth_in, :][:, cond_phi_in]
        sint_u_exp_in = sint_u_exp[cond_tth_in, :][:, cond_phi_in]
        int_d_exp_in = int_d_exp[cond_tth_in, :][:, cond_phi_in]
        sint_d_exp_in = sint_d_exp[cond_tth_in, :][:, cond_phi_in]

        proc, l_peak, l_refln, l_dd_out = self.calc_profile(
            tth_in, phi_in, l_crystal, l_peak_in=l_peak_in,
            l_refln_in=l_refln_in,
            l_refln_susceptibility_in=l_refln_susceptibility_in,
            l_dd_in=l_dd_in)
        proc.intensity_up = int_u_exp_in
        proc.intensity_up_sigma = sint_u_exp_in
        proc.intensity_down = int_d_exp_in
        proc.intensity_down_sigma = sint_d_exp_in

        # self.proc = proc
        # self.peaks = l_peak
        # self.reflns = l_refln
        self.dd = l_dd_out

        int_u_mod = proc.intensity_up_total
        int_d_mod = proc.intensity_down_total

        sint_sum_exp_in = (sint_u_exp_in**2 + sint_d_exp_in**2)**0.5

        chi_sq_u = ((int_u_mod-int_u_exp_in)/sint_u_exp_in)**2
        chi_sq_d = ((int_d_mod-int_d_exp_in)/sint_d_exp_in)**2
        chi_sq_sum = ((int_u_mod+int_d_mod-int_u_exp_in-int_d_exp_in) /
                      sint_sum_exp_in)**2
        chi_sq_dif = ((int_u_mod-int_d_mod-int_u_exp_in+int_d_exp_in) /
                      sint_sum_exp_in)**2

        cond_u = numpy.logical_not(numpy.isnan(chi_sq_u))
        cond_d = numpy.logical_not(numpy.isnan(chi_sq_d))
        cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))
        cond_dif = numpy.logical_not(numpy.isnan(chi_sq_dif))

        # exclude region
        try:
            exclude = self.exclude
            l_excl_tth_min = exclude.numpy_ttheta_min
            l_excl_tth_max = exclude.numpy_ttheta_max
            l_excl_phi_min = exclude.numpy_phi_min
            l_excl_phi_max = exclude.numpy_phi_max
            for excl_tth_min, excl_tth_max, excl_phi_min, excl_phi_max in \
                zip(l_excl_tth_min, l_excl_tth_max, l_excl_phi_min,
                    l_excl_phi_max):
                cond_1 = numpy.logical_or(tth_in < 1.*excl_tth_min,
                                          tth_in > 1.*excl_tth_max)
                cond_2 = numpy.logical_or(phi_in < 1.*excl_phi_min,
                                          phi_in > 1.*excl_phi_max)
                cond_11, cond_22 = numpy.meshgrid(cond_1, cond_2,
                                                  indexing="ij")
                cond_12 = numpy.logical_or(cond_11, cond_22)
                cond_u = numpy.logical_and(cond_u, cond_12)
                cond_d = numpy.logical_and(cond_d, cond_12)
                cond_sum = numpy.logical_and(cond_sum, cond_12)
        except AttributeError:
            pass

        chi_sq_u_val = (chi_sq_u[cond_u]).sum()
        n_u = cond_u.sum()

        chi_sq_d_val = (chi_sq_d[cond_d]).sum()
        n_d = cond_d.sum()

        chi_sq_sum_val = (chi_sq_sum[cond_sum]).sum()
        n_sum = cond_sum.sum()

        chi_sq_dif_val = (chi_sq_dif[cond_dif]).sum()
        n_dif = cond_dif.sum()
        chi2 = self.chi2
        flag_u = chi2.up
        flag_d = chi2.down
        flag_sum = chi2.sum
        flag_dif = chi2.diff

        chi_sq_val = (int(flag_u)*chi_sq_u_val + int(flag_d)*chi_sq_d_val +
                      int(flag_sum)*chi_sq_sum_val +
                      int(flag_dif)*chi_sq_dif_val)
        n = (int(flag_u)*n_u + int(flag_d)*n_d + int(flag_sum)*n_sum +
             int(flag_dif)*n_dif)
        # print(f"chi_sq_val/n: {chi_sq_val/n:.2f}    \
        #     chi_sq_val: {chi_sq_val: .2f}")
        # d_exp_out = {"chi_sq_val": chi_sq_val, "n": n}
        # d_exp_out.update(d_exp_prof_out)

        if flag_internal:
            refine_ls = RefineLs(number_reflns=n,
                                 goodness_of_fit_all=chi_sq_val/float(n),
                                 weighting_scheme="sigma")
            self.refine_ls = refine_ls
            proc.form_ttheta_phi_intensity_bkg_calc()
            proc.form_ttheta_phi_intensity_up()
            proc.form_ttheta_phi_intensity_up_sigma()
            proc.form_ttheta_phi_intensity_down()
            proc.form_ttheta_phi_intensity_down_sigma()
        return chi_sq_val, n

    def calc_for_iint(self, index_h, index_k, index_l, crystal,
                      flag_internal: bool = True):
        """Calculate the integral intensity for h, k, l reflections."""
        setup = self.setup
        field = float(setup.field)

        refln = crystal.calc_refln(index_h, index_k, index_l)

        refln_s = crystal.calc_refln_susceptibility(index_h, index_k, index_l)

        f_nucl = refln.numpy_f_calc

        sft_11 = refln_s.numpy_chi_11_calc
        sft_12 = refln_s.numpy_chi_12_calc
        sft_13 = refln_s.numpy_chi_13_calc
        sft_21 = refln_s.numpy_chi_21_calc
        sft_22 = refln_s.numpy_chi_22_calc
        sft_23 = refln_s.numpy_chi_23_calc
        sft_31 = refln_s.numpy_chi_31_calc
        sft_32 = refln_s.numpy_chi_32_calc
        sft_33 = refln_s.numpy_chi_33_calc

        sftm_11 = refln_s.numpy_moment_11_calc
        sftm_12 = refln_s.numpy_moment_12_calc
        sftm_13 = refln_s.numpy_moment_13_calc
        sftm_21 = refln_s.numpy_moment_21_calc
        sftm_22 = refln_s.numpy_moment_22_calc
        sftm_23 = refln_s.numpy_moment_23_calc
        sftm_31 = refln_s.numpy_moment_31_calc
        sftm_32 = refln_s.numpy_moment_32_calc
        sftm_33 = refln_s.numpy_moment_33_calc

        _11, _12 = sftm_11+field*sft_11, sftm_12+field*sft_12
        _21, _13 = sftm_21+field*sft_21, sftm_13+field*sft_13
        _22, _23 = sftm_22+field*sft_22, sftm_23+field*sft_23
        _31, _32 = sftm_31+field*sft_31, sftm_32+field*sft_32
        _33 = sftm_33+field*sft_33
        _ij = (_11, _12, _13, _21, _22, _23, _31, _32, _33)
        cell = crystal.cell
        # k_loc = cell.calc_k_loc(index_h, index_k, index_l)
        t_ij = cell.calc_m_t(index_h, index_k, index_l)
        # FIXME: I would like to recheck the expression for T
        #        and expression SIGMA = T^T CHI T

        t_tr_ij = (t_ij[0], t_ij[3], t_ij[6],
                   t_ij[1], t_ij[4], t_ij[7],
                   t_ij[2], t_ij[5], t_ij[8])
        th_11, th_12, th_13, th_21, th_22, th_23, th_31, th_32, th_33 = \
            calc_mRmCmRT(t_tr_ij, _ij)

        # f_m_p_sin_sq = (field**2)*abs(0.5*(th_11*th_11.conjugate()+\
        #                th_22*th_22.conjugate())+th_12*th_12.conjugate())
        # f_m_p_cos_sq = (field**2)*abs(th_13*th_13.conjugate()+\
        #                 th_23*th_23.conjugate())
        # f_m_p_field = 0.5*field*(th_11+th_22)

        f_nucl_sq = abs(f_nucl*f_nucl.conjugate())
        f_m_p_sin_sq = abs(0.5 * (th_11 * th_11.conjugate()+th_22 *
                                  th_22.conjugate())+th_12 * th_12.conjugate())
        f_m_p_cos_sq = abs(th_13 * th_13.conjugate() +
                           th_23 * th_23.conjugate())
        f_m_p_field = 0.5 * (th_11+th_22)
        cross_sin = 2. * (f_nucl.real * f_m_p_field.real +
                          f_nucl.imag * f_m_p_field.imag)

        return f_nucl_sq, f_m_p_sin_sq, f_m_p_cos_sq, cross_sin, refln, refln_s

    def _gauss_pd(self, tth_2d):
        """One dimensional gauss powder diffraction."""
        ag, bg = self.ag, self.bg
        val_1 = bg*tth_2d**2
        val_2 = numpy.where(val_1 < 5., numpy.exp(-val_1), 0.)
        self.gauss_pd = ag*val_2

    def _lor_pd(self, tth_2d):
        """One dimensional lorentz powder diffraction."""
        al, bl = self.al, self.bl
        self.lor_pd = al*1./(1.+bl*tth_2d**2)

    def calc_shape_profile(
            self, tth, phi, tth_hkl, phase_igsize: float = 0.,
            phase_u: float = 0., phase_v: float = 0., phase_w: float = 0.,
            phase_x: float = 0., phase_y: float = 0.):
        """
        Calculate profile in the range ttheta.

        For reflections placed on
        ttheta_hkl with i_g parameter by default equal to zero

        tth, phi, tth_hkl in degrees
        """
        setup = self.setup
        zero_shift = float(setup.offset_ttheta)
        tth_zs = tth-zero_shift

        resolution = self.pd2d_instr_resolution

        h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = resolution.calc_resolution(
            tth_hkl, phase_igsize=phase_igsize, phase_u=phase_u,
            phase_v=phase_v, phase_w=phase_w, phase_x=phase_x, phase_y=phase_y)

        tth_3d, phi_3d, tth_hkl_3d = numpy.meshgrid(tth_zs, phi, tth_hkl,
                                                    indexing="ij")

        self.ag = numpy.meshgrid(tth_zs, phi, a_g, indexing="ij")[2]
        self.bg = numpy.meshgrid(tth_zs, phi, b_g, indexing="ij")[2]
        self.al = numpy.meshgrid(tth_zs, phi, a_l, indexing="ij")[2]
        self.bl = numpy.meshgrid(tth_zs, phi, b_l, indexing="ij")[2]
        eta_3d = numpy.meshgrid(tth_zs, phi, eta, indexing="ij")[2]
        self.eta = eta_3d

        self._gauss_pd(tth_3d-tth_hkl_3d)
        self._lor_pd(tth_3d-tth_hkl_3d)
        g_pd2d_3d = self.gauss_pd
        l_pd2d_3d = self.lor_pd

        np_shape_3d = eta_3d * l_pd2d_3d + (1.-eta_3d) * g_pd2d_3d

        asymmetry = self.pd2d_instr_reflex_asymmetry
        np_ass_2d = asymmetry.calc_asymmetry(tth_zs, tth_hkl, h_pv)
        np_ass_3d = np_ass_2d[:, numpy.newaxis, :] * numpy.ones(
            phi.size, dtype=float)[numpy.newaxis, :, numpy.newaxis]

        # Lorentz factor
        tth_rad = tth_zs*numpy.pi/180.
        np_lor_1d = 1./(numpy.sin(tth_rad)*numpy.sin(0.5*tth_rad))
        np_lor_3d = numpy.meshgrid(np_lor_1d, phi, tth_hkl, indexing="ij")[0]

        profile_3d = np_shape_3d*np_ass_3d*np_lor_3d

        return profile_3d, tth_zs, h_pv

    def params_to_cif(self, separator="_", flag: bool = False,
                      flag_minimal: bool = True) -> str:
        """Save parameters to cif format."""
        ls_out = []
        l_cls = (Pd2dBackground, Pd2dInstrResolution, PhaseL, DiffrnRadiation,
                 Setup, Range, Chi2, Extinction, ExcludeL,
                 Pd2dInstrReflexAsymmetry, Texture)
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def data_to_cif(self, separator="_", flag: bool = False,
                    flag_minimal: bool = True) -> str:
        """Save data to cif format."""
        ls_out = []
        l_cls = (Pd2dMeas, )
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def calc_to_cif(self, separator="_", flag=False, flag_minimal=True) -> str:
        """Save calculations to cif format."""
        ls_out = []
        l_cls = (RefineLs, ReflnL, ReflnSusceptibilityL, Pd2dPeakL, Pd2dProc)
        l_obj = [item for item in self.items if type(item) in l_cls]
        l_itemn = [item for item in l_obj if isinstance(item, ItemN)]
        l_loopn = [item for item in l_obj if isinstance(item, LoopN)]
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_itemn])
        ls_out.extend([_.to_cif(separator=separator)+"\n" for _ in l_loopn])
        return "\n".join(ls_out)

    def apply_constraints(self):
        """Apply constraints."""
        pass

# s_cont = """
#  data_powder2d
#  _setup_wavelength     0.840
#  _setup_field          1.000
#  _setup_offset_2theta -0.385
#  _chi2_sum  True
#  _chi2_diff False
#  _chi2_up   False
#  _chi2_down False
#  _range_2theta_min     4.000
#  _range_2theta_max    80.000
#  _range_phi_min       -2.000
#  _range_phi_max       40.000
#  _pd2d_background_2theta_phi_intensity
#  ;
#  2 5 80
#  5 2. 2.(1)
#  40 2. 2.
#  ;
#  loop_
#  _exclude_2theta_min
#  _exclude_2theta_max
#  _exclude_phi_min
#  _exclude_phi_max
#  0.0 1.0 0. 1.0
#  _pd2d_instr_reflex_asymmetry_p1 0.0
#  _pd2d_instr_reflex_asymmetry_p2 0.0
#  _pd2d_instr_reflex_asymmetry_p3 0.0
#  _pd2d_instr_reflex_asymmetry_p4 0.0
#  _diffrn_radiation_polarization 0.0
#  _diffrn_radiation_efficiency   1.0

#  _pd2d_instr_resolution_u 16.9776
#  _pd2d_instr_resolution_v -2.8357(1)
#  _pd2d_instr_resolution_w  0.5763
#  _pd2d_instr_resolution_x  0.0
#  _pd2d_instr_resolution_y  0.0
#  loop_
#  _phase_label
#  _phase_scale
#  _phase_igsize
#  Fe3O4 0.02381() 0.0
#  _pd2d_meas_2theta_phi_intensity_up
#  ;
#  2 5 80
#  5 2. 2.
#  40 2. 2.
#  ;
#  _pd2d_meas_2theta_phi_intensity_up_sigma
#  ;
#  2 5 80
#  5 2. 2.
#  40 2. 2.
#  ;
#  _pd2d_meas_2theta_phi_intensity_down
#  ;
#  2 5 80
#  5 2. 2.
#  40 2. 2.
#  ;
#  _pd2d_meas_2theta_phi_intensity_down_sigma
#  ;
#  2 5 80
#  5 2. 2.
#  40 2. 2.
#  ;
# """
# obj = Pd2d.from_cif(s_cont)
# for item in obj.items:
#     print(item, end="\n\n")
# print(50*"*", "\n", obj.is_defined())
# for var_name in obj.get_variable_names():
#     print(var_name)



"""Setup and SetupL classes."""
from typing import NoReturn
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class Setup(ItemN):
    """Setup class.

    Attributes
    ----------
        - wavelength (mandatory) (in Angstrems)
        - field (optional) (in Tesla)
        - offset_ttheta (optional for powder 1d and 2d) (in degrees)
        - offset_phi (optional for powder 2d) (in degrees)
        - ratio_lambdaover2 (optional, for single diffraction)
    """

    ATTR_MANDATORY_NAMES = ("wavelength", )
    ATTR_MANDATORY_TYPES = (float, )
    ATTR_MANDATORY_CIF = ("wavelength", )

    ATTR_OPTIONAL_NAMES = ("field", "offset_ttheta", "offset_phi",
                           "ratio_lambdaover2")
    ATTR_OPTIONAL_TYPES = (float, float, float, float)
    ATTR_OPTIONAL_CIF = ("field", "offset_2theta", "offset_phi",
                         "ratio_lambda/2")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("wavelength", "offset_ttheta", "offset_phi",
                "ratio_lambdaover2")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {'wavelength': "{:.4f}", 'field': "{:.2f}",
                 'offset_ttheta': "{:.3f}", 'offset_phi': "{:.3f}",
                 "ratio_lambdaover2": "{:.3f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {"offset_2theta": 0.}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "setup"

    def __init__(self, **kwargs) -> NoReturn:
        super(Setup, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"wavelength": 0., "ratio_lambdaover2": 0.}

        # defined for ani integer and float parameters
        D_MAX = {"ratio_lambdaover2": 1.}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class SetupL(LoopN):
    """Description of Setup in loop."""

    ITEM_CLASS = Setup
    ATTR_INDEX = None

    def __init__(self, loop_name=None) -> NoReturn:
        super(SetupL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name

# s_cont = """
#   loop_
#  _setup_wavelength   0.84
#  _setup_field        1.00
#  _setup_offset_2theta -0.385
#  _setup_offset_phi -0.385
#     0.84 1.0 -0.385 -0.385
#     1.0  1.5 0.7 0.3
#   """

# obj = SetupL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj[0], end="\n\n")



import numpy
from typing import NoReturn
from cryspy.A_functions_base.function_1_matrices import\
    calc_product_matrices, calc_product_matrix_vector
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.C_item_loop_classes.cl_1_cell import Cell


class PdProc(ItemN):
    """
    This section contains the diffraction data set after processing
    and application of correction terms. If the data set is
    reprocessed, this section may be replaced (with the addition of
    a new _pd_block_id entry).

    Mandatory attributes:
        - ub_11, ub_12, ub_13, ub_21, ub_22, ub_23, ub_31, ub_32, ub_33

    Optional attributes:
        - occupancy
        - adp_type
        - u_iso_or_equiv
        - u_equiv_geom_mean
        - b_iso_or_equiv
        - multiplicity
        - wyckoff_symbol
        - cartn_x
        - cartn_y
        - cartn_z

    Internal attributes:
        - scat_length_neutron

    Internal protected attributes:
        - space_group_wyckoff
        - constr_number
    """
    ATTR_MANDATORY_NAMES = ("ttheta", )
    ATTR_MANDATORY_TYPES = (float, )
    ATTR_MANDATORY_CIF = ("2theta", )

    ATTR_OPTIONAL_NAMES = (
        "ttheta_corrected", "d_spacing", "intensity_up_net",
        "intensity_down_net", "intensity_up_total", "intensity_down_total", 
        "intensity_bkg_calc", "intensity_net", "intensity_total",
        "intensity_diff_total", "intensity_up", "intensity_up_sigma",
        "intensity_down", "intensity_down_sigma", "intensity",
        "intensity_sigma")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float, float, float,
                           float, float, float, float, float, float, float, 
                           float, float)
    ATTR_OPTIONAL_CIF = (
        "2theta_corrected", "d_spacing", "intensity_up_net",
        "intensity_down_net", "intensity_up_total", "intensity_down_total", 
        "intensity_bkg_calc", "intensity_net", "intensity_total",
        "intensity_diff_total", "intensity_up", "intensity_up_sigma",
        "intensity_down", "intensity_down_sigma", "intensity",
        "intensity_sigma")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {
        "ttheta_corrected": "{:.2f}", "d_spacing": "{:.5f}",
        "intensity_up_net": "{:.2f}", "intensity_down_net": "{:.2f}",
        "intensity_up_total": "{:.2f}", "intensity_down_total": "{:.2f}",
        "intensity_bkg_calc": "{:.2f}", "intensity_net": "{:.2f}",
        "intensity_total": "{:.2f}", "intensity_diff_total": "{:.2f}",
        "intensity_down": "{:.2f}", "intensity_down_sigma": "{:.2f}",
        "intensity": "{:.2f}", "intensity_sigma": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "pd_proc"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdProc, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class PdProcL(LoopN):
    """
    This section contains the diffraction data set after processing
    and application of correction terms. If the data set is
    reprocessed, this section may be replaced (with the addition of
    a new _pd_block_id entry).

    """
    ITEM_CLASS = PdProc
    ATTR_INDEX = "ttheta"
    def __init__(self, loop_name = None) -> NoReturn:
        super(PdProcL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name
   

# s_cont = """
#  loop_
#  _pd_proc_ttheta
#  _pd_proc_ttheta_corrected
#  _pd_proc_d_spacing
#  _pd_proc_intensity_up_net
#  _pd_proc_intensity_down_net
#  _pd_proc_intensity_up_total
#  _pd_proc_intensity_down_total
#  _pd_proc_intensity_bkg_calc
#  _pd_proc_intensity_up
#  _pd_proc_intensity_up_sigma
#  _pd_proc_intensity_down
#  _pd_proc_intensity_down_sigma
#   4.00  3.9  11.2   400.00   317.00   460.00   377.00   60.00    465.80000   128.97000   301.88000   129.30000"""

# obj = PdProcL.from_cif(s_cont)
# print(obj, end="\n\n")



"""
Functions:
    - transs
    - calc_GCF

"""
import numpy
import math


def transs(lmax: int, nn: int, zeta: float, sthovl: numpy.ndarray):
    """
    Calculate integral.

    $$
    ( r**nn * exp(-zeta r) * j_l(Qr) r**2 dr) / Q**l
    $$

    for given l

    Q = 4 pi sint / lambda

    """
    Q = 4. * math.pi * sthovl
    a = [0.*Q for h in range(17)]
    ff = numpy.zeros(shape=(sthovl.size, lmax+1), dtype=float)
    d = Q**2+zeta**2
    a[0] = 0.0*Q
    a[1] = 1./d
    n = nn-1
    tz = 2.*zeta
    ts = 4.*math.pi
    for l in range(lmax+1):
        ll = l+1
        if (ll != 1):
            a[ll] = a[ll-1] * ts * l * 1./d
            a[ll-1] = 0.0*Q
        for nx in range(ll, n+1):
            i1 = nx
            i2 = nx+1
            i3 = i2+1
            a[i3-1] = (tz*nx*a[i2-1] - (nx+l)*(nx-ll)*a[i1-1]) * 1./d
        ff[:, ll-1] = a[i3-1]
    return ff

def calc_GCF(ln1: list, ldzeta1: list, lcoeff1: list, kappa1: float,
             ln2: list, ldzeta2: list, lcoeff2: list, kappa2: float,
             sthovl: float, lmax: float):
    GCF = numpy.zeros(shape=(sthovl.size, lmax+1), dtype=float)
    for n1, dzeta1, coeff1 in zip(ln1, ldzeta1, lcoeff1):
        # transformation from atomic units to inverse angstrems
        zetaA1 = dzeta1*kappa1/0.529177 
        Norm1 = math.sqrt(((2.*zetaA1)**(2*n1+1))/math.factorial(2*n1))
        for n2, dzeta2, coeff2 in zip(ln2, ldzeta2, lcoeff2):
            # transformation from atomic units to inverse angstrems
            zetaA2 = dzeta2*kappa2/0.529177
            Norm2 = math.sqrt(((2.*zetaA2)**(2*n2+1))/math.factorial(2*n2))

            nn, zeta = (n1+n2), (zetaA1+zetaA2)
            GCF += coeff1*coeff2*Norm1*Norm2*transs(lmax, nn, zeta, sthovl)
    return GCF


FUNCTIONS = [
    transs,
    calc_GCF
]



from typing import NoReturn
import math
import numpy
from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class DiffrnRefln(ItemN):
    """
    DiffrnRefln class.

    Data items in the DIFFRN_REFLN category record details about
    the intensities measured in the diffraction experiment.

    The DIFFRN_REFLN data items refer to individual intensity
    measurements and must be included in looped lists.

    Attributes:
        - 

    """
    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = ("fr", "fr_sigma", "fr_calc", "intensity_up_calc",
                           "intensity_down_calc")
    ATTR_OPTIONAL_TYPES = (float, float, float, float, float)
    ATTR_OPTIONAL_CIF = ("fr", "fr_sigma", "fr_calc" , "intensity_up_calc",
                         "intensity_down_calc")

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"fr_calc": "{:.5f}", "intensity_up_calc": "{:.2f}",
                 "intensity_down_calc": "{:.2f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False

    PREFIX = "diffrn_refln"

    def __init__(self, **kwargs) -> NoReturn:
        super(DiffrnRefln, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"fr": 0., "fr_sigma": 0., "fr_calc": 0.,
                 "intensity_up_calc": 0., "intensity_down_calc": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class DiffrnReflnL(LoopN):
    """
    Description of chi2 in loop.

    """
    ITEM_CLASS = DiffrnRefln
    ATTR_INDEX = None
    def __init__(self, loop_name = None) -> NoReturn:
        super(DiffrnReflnL, self).__init__()
        self.__dict__["items"] = []
        self.__dict__["loop_name"] = loop_name

    def report_agreement_factor_exp(self):
        """
        Make a report about experimental agreement factor in string format.
        """

        l_chi_sq_exp, l_ag_f_exp = [], []
        l_diff = []
        n_3s, n_2s, n_1s, n_0s = 0, 0, 0, 0
        l_hkl = [(int(_1), int(_2), int(_3)) for _1, _2, _3 in
                 zip(self.index_h, self.index_k, self.index_l)]
        for _hkl, _fr_1, _fr_sigma_1 in zip(l_hkl, self.fr, self.fr_sigma):
            _diff = abs(_fr_1-1.)/float(_fr_sigma_1)
            if _diff >= 3.:
                n_3s += 1
            elif _diff >= 2.:
                n_2s += 1
            elif _diff >= 1.:
                n_1s += 1
            else:
                n_0s += 1
            if _fr_1 <= 1.:
                l_diff.append(1./_fr_1-1.)
            else:
                l_diff.append(_fr_1-1.)

            _mhkl = (-1 * _hkl[0], -1 * _hkl[1], -1 * _hkl[2])
            if _mhkl in l_hkl:
                ind_mhkl = l_hkl.index(_mhkl)
                _fr_2, _fr_sigma_2 = self.fr[ind_mhkl], self.fr_sigma[ind_mhkl]
                _fr_sigma = 1. / (_fr_sigma_1 ** (-2) + _fr_sigma_2 ** (-2)) ** 0.5
                _fr_average = (_fr_1 * _fr_sigma_1 ** (-2) + _fr_2 * _fr_sigma_2 ** (-2)) * _fr_sigma ** 2
                delta_fr = abs(_fr_1 - _fr_average)
                chi_sq_exp = (delta_fr / _fr_sigma_1) ** 2
                l_chi_sq_exp.append(chi_sq_exp)
                if math.isclose(_fr_1-1., 0):
                    ag_f_exp = abs((_fr_1 - _fr_average) / (_fr_1 - 1.+_fr_sigma_1))
                else:
                    ag_f_exp = abs((_fr_1 - _fr_average) / (_fr_1 - 1.))
                l_ag_f_exp.append(ag_f_exp)
                # print("hkl: {:4} {:4} {:4}".format(_hkl[0], _hkl[1], _hkl[2]))
                # print("chi_sq_exp: {:.3f} ".format(chi_sq_exp))
                # print("ag_f_exp: {:.3f} ".format(ag_f_exp))
        ls_out = []
        n = n_0s + n_1s + n_2s + n_3s
        ls_out.append(f"\n Number of measured reflections: {n:4}")
        ls_out.append(f"      range of h is {min([_[0] for _ in l_hkl]):3},{max([_[0] for _ in l_hkl]):3} ")
        ls_out.append(f"               k is {min([_[1] for _ in l_hkl]):3},{max([_[1] for _ in l_hkl]):3} ")
        ls_out.append(f"               l is {min([_[2] for _ in l_hkl]):3},{max([_[2] for _ in l_hkl]):3} ")
        ls_out.append(f" max(FR_exp - 1) is {max(l_diff):5.3f} ")
        ls_out.append(f" N+1 > |FR_exp - 1|/FR_sigma > N: ")
        ls_out.append(f" |FR_exp - 1|/FR_sigma < 1: {n_0s:4}, {100*float(n_0s)/float(n):5.1f}%  ")
        ls_out.append(f"                     N = 1: {n_1s:4}, {100*float(n_1s)/float(n):5.1f}%  ")
        ls_out.append(f"                     N = 2: {n_2s:4}, {100 * float(n_2s) / float(n):5.1f}% ")
        ls_out.append(f" |FR_exp - 1|/FR_sigma > 3: {n_3s:4}, {100 * float(n_3s) / float(n):5.1f}% ")

        n_friedel = len(l_chi_sq_exp)
        ls_out.append(f"Total number of Friedel reflections is {n_friedel:}.")
        if n_friedel != 0:
            ls_out.append(f"  (|FR_exp-FR_av.|/|FR_sigma|)^2  is {sum(l_chi_sq_exp) / n_friedel:.2f}")
            ls_out.append(f"   |FR_exp-FR_av.|/|FR_exp-1| per reflection is {(100*sum(l_ag_f_exp)/n_friedel):.2f}% ")

        return "\n".join(ls_out)

    def calc_chi_sq_points(self) -> (float, int):
        """Calculate chi_sq and points if fr, fr_sigma, fr_calc are given."""
        l_fr, l_fr_sigma, l_fr_calc = self.fr, self.fr_sigma, self.fr_calc
        if (l_fr is None) & (l_fr_sigma is None) & (l_fr_calc is None):
            return (None, None)

        l_chi_sq = [(abs(float(fr-fr_calc)/float(fr_sigma))**2)
                    for fr, fr_sigma, fr_calc
                    in zip(l_fr, l_fr_sigma, l_fr_calc)]
        return sum(l_chi_sq), len(l_chi_sq)

    def report_chi_sq_exp(self, cell=None) -> str:
        """
        Make a report about experimental chi_sq in string format.

        cell is unit cell object.
        """
        ls_out = []
        l_hkl = [(_1, _2, _3) for _1, _2, _3 in zip(self.index_h, self.index_k, self.index_l)]
        l_fr, l_fr_sigma, l_fr_calc = self.fr, self.fr_sigma, self.fr_calc
        if (l_fr is None) & (l_fr_sigma is None) & (l_fr_calc is None):
            return "\n".join(ls_out)
        n = len(l_fr)
        n_1s, n_2s, n_3s = 0, 0, 0
        l_chi_sq, l_worsest, l_af_f, l_af_r = [], [], [], []
        for _hkl, _fr, _fr_sigma, _fr_calc in zip(l_hkl, l_fr, l_fr_sigma, l_fr_calc):
            _diff = abs(float(_fr-_fr_calc)/float(_fr_sigma))
            l_chi_sq.append(_diff**2)
            l_af_f.append(abs(float(_fr - _fr_calc) / float(_fr)))
            if _fr != 1.:
                l_af_r.append(abs(float(_fr-_fr_calc)/float(_fr-1)))
            if _diff <= 1.:
                n_1s +=1
            elif _diff <= 2.:
                n_2s += 1
            elif _diff <= 3.:
                n_3s += 1
            else:
                l_worsest.append((_hkl, _fr, _fr_sigma, _fr_calc, _diff))
        ls_out.append(f"Total number of reflections is {n:}.")
        ls_out.append(f"  (|FR_exp-FR_mod|/|FR_sigma|)^2 per reflection is {sum(l_chi_sq)/float(n):.2f}")
        ls_out.append(f"   |FR_exp-FR_mod|/|FR_exp|      per reflection is {100*sum(l_af_f) / float(n):.2f}%")
        ls_out.append(f"   |FR_exp-FR_mod|/|FR_exp-1|    per reflection is {100*sum(l_af_r) / float(n):.2f}%")
        ls_out.append(f"           (reflections with FR_exp = 1 are excluded)")
        n_worsest = len(l_worsest)
        ls_out.append(f"\nReflections in range  ")
        ls_out.append(f" (N-1)*FR_sigma < |FR_exp - FR_mod| < N*FR_sigma: ")
        ls_out.append(f"      N = 1: {n_1s:}/{n:} ={100*float(n_1s)/float(n):5.1f}% ({2*34.1:4.1f}%, three sigma rule) ")
        ls_out.append(f"      N = 2: {n_2s:}/{n:} ={100 * float(n_2s) / float(n):5.1f}% ({2*(13.6):4.1f}%, three sigma rule)")
        ls_out.append(f"      N = 3: {n_3s:}/{n:} ={100 * float(n_3s) / float(n):5.1f}% ({2*(2.1):4.1f}%, three sigma rule)")
        ls_out.append(f"      N > 3: {n_worsest:}/{n:} ={100 * float(n_worsest) / float(n):5.1f}% ({2*(0.1):4.1f}%, three sigma rule)")
        l_worsest.sort(key=lambda x: x[4], reverse=True)
        if len(l_worsest) > 1:
            if n_worsest > 10: n_worsest = 10
            ls_out.append("\nThe ten worsest reflections:")
            ls_out.append("  h  k  l       FR FR_sigma  FR_calc  diff")
            for (_hkl, _fr, _fr_sigma, _fr_calc, _diff) in l_worsest[:n_worsest]:
                ls_out.append(f"{_hkl[0]:3}{_hkl[1]:3}{_hkl[2]:3}{_fr:9.5f}{_fr_sigma:9.5f}{_fr_calc:9.5f}{_diff:6.1f}")
        if cell is not None:
            np_h = numpy.array(self.index_h, dtype=int)
            np_k = numpy.array(self.index_k, dtype=int)
            np_l = numpy.array(self.index_l, dtype=int)
            np_sthovl = cell.calc_sthovl(index_h=np_h, index_k=np_k,
                                         index_l=np_l)
            np_fr, np_fr_sigma = numpy.array(l_fr, dtype=float), numpy.array(l_fr_sigma, dtype=float) 
            np_fr_calc = numpy.array(l_fr_calc, dtype=float)
            np_diff = numpy.abs((np_fr-np_fr_calc)/np_fr_sigma)
            n_bins = 10
            np_val, np_sthovl_bins = numpy.histogram(np_sthovl, bins=n_bins)
            ls_out.append(f"\nDistribution of reflection in sin(theta)/lambda range")
            ls_out.append(f"sthovl_1 sthovl_2 <Exp-Mod/Sigma> n_points")
            for _1, _2, _3 in zip(np_sthovl_bins[:-1], np_sthovl_bins[1:], np_val):
                np_flag = numpy.logical_and(np_sthovl >= _1, np_sthovl < _2) 
                res = np_diff[np_flag]
                if res.size == 0:
                    ls_out.append(f"{_1:8.3f} {_2:8.3f}:       None          0")
                else:
                    ls_out.append(f"{_1:8.3f} {_2:8.3f}:    {np_diff[np_flag].mean():7.3f} {_3:10}")
            n_vals = numpy.histogram(np_sthovl, bins=n_bins)
        return "\n".join(ls_out)

# s_cont = """
#   loop_
#   _diffrn_refln_index_h
#   _diffrn_refln_index_k
#   _diffrn_refln_index_l
#   _diffrn_refln_fr
#   _diffrn_refln_fr_sigma
#       0    0    8   0.64545   0.01329 
#       2    0    6   1.75682   0.0454
# """

# obj = DiffrnReflnL.from_cif(s_cont)
# print(obj, end="\n\n")
# print(obj.report_agreement_factor_exp(), end="\n\n")
# print(obj.numpy_fr_sigma, end="\n\n")



"""Description of RhoChi class."""
__author__ = 'ikibalin'
__version__ = "2020_08_19"
import os
import os.path
from warnings import warn
import numpy
from typing import NoReturn, List
import scipy
import scipy.optimize

from cryspy.A_functions_base.function_1_error_simplex import \
    error_estimation_simplex

from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.E_data_classes.cl_1_crystal import Crystal
from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn
from cryspy.E_data_classes.cl_2_pd import Pd
from cryspy.E_data_classes.cl_2_pd2d import Pd2d


class RhoChi(GlobalN):
    """
    Class to describe RhoChi class.

    Attributes
    ----------
        - crystal_#name (mandatory)
        - diffrn_#name
        - pd_#name
        - pd2d_#name

    Methods
    -------
        - crystals()
        - experiments()
        - refine()
        - calc_chi_sq
        - params_to_cif
        - data_to_cif
        - calc_to_cif
        - estimate_FM

    """

    CLASSES_MANDATORY = ()
    CLASSES_OPTIONAL = (Crystal, MagCrystal, Diffrn, Pd, Pd2d)
    # CLASSES_INTERNAL = ()

    CLASSES = CLASSES_MANDATORY + CLASSES_OPTIONAL

    PREFIX = "rhochi"

    # default values for the parameters
    D_DEFAULT = {}

    def __init__(self, global_name=None, **kwargs) -> NoReturn:
        super(RhoChi, self).__init__()

        self.__dict__["items"] = []
        self.__dict__["global_name"] = global_name

        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def form_object(self):
        """Form object."""
        self.apply_constraint()
        for item in self.items:
            if item.is_defined():
                item.form_object()

    def apply_constraint(self):
        """Apply constraints."""
        for item in self.items:
            item.apply_constraints()

    def experiments(self):
        """List of expreiments."""
        return [item for item in self.items if isinstance(item, (Diffrn, Pd,
                                                                 Pd2d))]

    def crystals(self):
        """List of crystals."""
        return [item for item in self.items if isinstance(item, (Crystal,
                                                                 MagCrystal))]

    def calc_chi_sq(self, flag_internal=True):
        """
        Calculate chi square.

        Keyword Arguments
        -----------------
            - flag_internal: a flag to calculate internal objects
              (default is True)

        Output arguments
        ----------------
            - chi_sq_val: chi square of flip ratio
              (Sum_i ((y_e_i - y_m_i) / sigma_i)**2)
            - n: number of measured reflections
        """
        self.apply_constraint()

        l_crystal = self.crystals()

        chi_sq_res, n_res = 0., 0.
        for experiment in self.experiments():
            chi_sq, n = experiment.calc_chi_sq(l_crystal,
                                               flag_internal=flag_internal)
            experiment.chi_sq = chi_sq
            experiment.n = n
            chi_sq_res += chi_sq
            n_res += n
        return chi_sq_res, n_res

    def refine(self, optimization_method: str = "BFGS", disp: bool = False,
               d_info: dict = None):
        """
        Minimization procedure.

        Accesible parameters for optimization_method
        --------------------------------------------
            - "BFGS" (default)
            - "simplex"
            - "basinhopping"
        """

        if d_info is not None:
            d_info_keys = d_info.keys()
            if "stop" not in d_info_keys:
                d_info["stop"] = False
            d_info["print"] = "Refinement is running."

        flag = True

        # self.remove_internal_objs

        self.apply_constraint()
        l_var_name = self.get_variable_names()

        if len(l_var_name) == 0:
            chi_sq, n = self.calc_chi_sq()
            # self._show_message(f"chi_sq/n {chi_sq/n:.2f} (n = {int(n):}).")
            dict_out = {"flag": flag, "res": None, "chi_sq": chi_sq, "n": n}
            return dict_out

        val_0 = numpy.array([self.get_variable_by_name(var_name)
                             for var_name in l_var_name], dtype=float)

        sign = 2*(numpy.array(val_0 >= 0., dtype=int)-0.5)
        param_0 = numpy.log(abs(val_0)*(numpy.e-1.)+1.)*sign
        coeff_norm = numpy.where(val_0 == 0., 1., val_0) / \
            numpy.where(param_0 == 0., 1., param_0)
        hes_coeff_norm = numpy.matmul(coeff_norm[:, numpy.newaxis],
                                      coeff_norm[numpy.newaxis, :])

        chi_sq, n = self.calc_chi_sq(flag_internal=True)

        def tempfunc(l_param):
            for var_name, param, coeff in zip(l_var_name, l_param, coeff_norm):
                self.set_variable_by_name(var_name, param*coeff)
            chi_sq, n_points = self.calc_chi_sq(flag_internal=False)
            if n_points < n:
                res_out = 1.0e+308
            else:
                res_out = (chi_sq*1./float(n_points))
            return res_out

        if optimization_method == "basinhopping":
            # basinhopping
            res = scipy.optimize.basinhopping(
                tempfunc, param_0, niter=10, T=10, stepsize=0.1, interval=20,
                disp=disp)
            _dict_out = {"flag": flag, "res": res}
        elif optimization_method == "simplex":
            # simplex
            res = scipy.optimize.minimize(
                tempfunc, param_0, method='Nelder-Mead',
                callback=lambda x: self._f_callback(
                    disp, coeff_norm, x, param_name=l_var_name, d_info=d_info),
                options={"fatol": 0.01*n})

            m_error, dist_hh = error_estimation_simplex(
                res["final_simplex"][0], res["final_simplex"][1], tempfunc)

            l_sigma = []
            for i, val_2 in zip(range(m_error.shape[0]), dist_hh):
                # slightly change definition, instead of (n-k) here is n
                error = (abs(m_error[i, i])*1./n)**0.5
                if m_error[i, i] < 0.:
                    warn("Negative diagonal elements of Hessian.", UserWarning)
                if val_2 > error:
                    warn("Minimum is not found.", UserWarning)

                l_sigma.append(max(error, val_2))

            for var_name, sigma, coeff in \
                    zip(l_var_name, l_sigma, coeff_norm):
                hh = tuple((f"{var_name[-1][0]:}_sigma", ) + var_name[-1][1:])
                var_name_sigma = tuple(var_name[:-1]+(hh, ))
                self.set_variable_by_name(var_name_sigma, sigma*coeff)

            _dict_out = {"flag": flag, "res": res}
        else:
            # BFGS
            res = scipy.optimize.minimize(
                tempfunc, param_0, method='BFGS',
                callback=lambda x: self._f_callback(
                    disp, coeff_norm, x, param_name=l_var_name, d_info=d_info),
                options={"disp": disp})

            _dict_out = {"flag": flag, "res": res}

            hess_inv = res["hess_inv"] * hes_coeff_norm
            l_param = res["x"]
            l_sigma = (abs(numpy.diag(hess_inv)*1./float(n)))**0.5
            for var_name, sigma, param, coeff in \
                    zip(l_var_name, l_sigma, l_param, coeff_norm):
                hh = tuple((f"{var_name[-1][0]:}_sigma", ) + var_name[-1][1:])
                var_name_sigma = tuple(var_name[:-1]+(hh, ))
                self.set_variable_by_name(var_name_sigma, sigma)
                self.set_variable_by_name(var_name, param*coeff)

        chi_sq, n = self.calc_chi_sq(flag_internal=True)

        return _dict_out

    def _f_callback(self, *arg, param_name: List[tuple] = None,
                    d_info: dict = None) -> bool:
        flag_out = False
        disp = arg[0]
        if disp:
            coeff_norm = arg[1]
            res_x = arg[2]
            ls_out = " ".join(["{:12.5f}".format(_1*_2)
                               for _1, _2 in zip(res_x, coeff_norm)])
            print(ls_out)

        if d_info is not None:
            coeff_norm = arg[1]
            l_param = arg[2]
            if param_name is not None:
                ls_out = [f"{name[-1][0]:}: {coeff*param: 10.5f}" for name,
                          param, coeff in zip(param_name, l_param, coeff_norm)]
            else:
                ls_out = [f"{coeff*param: 10.5f}" for param, coeff in zip(
                    l_param, coeff_norm)]
            d_info["print"] = "Best solution:\n\n"+"\n".join(ls_out)
            flag_out = d_info["stop"]
        return flag_out

    def save_to_file(self, f_name):
        """Save to file."""
        self.file_input = f_name
        if os.path.basename(f_name) == "main.rcif":
            self.save_to_files()
        else:
            with open(f_name, "w") as fid:
                fid.write(self.to_cif())

    def save_to_files(self):
        """Save to files."""
        if self.file_input is None:
            f_dir = "."
        else:
            f_dir = os.path.dirname(self.file_input)
        f_main = os.path.join(f_dir, "main.rcif")
        ls_main = []
        ls_main.append("global_{:}\n".format(self.global_name))
        for experiment in self.experiments():
            ls_main.append(f"_add_url {experiment.data_name:}_data.rcif\n")
            ls_main.append(f"_add_url {experiment.data_name:}_calc.rcif\n")
        for crystal in self.crystals():
            ls_main.append("\n"+crystal.to_cif())
        for experiment in self.experiments():
            ls_main.append("\ndata_{:}".format(experiment.data_name))
            ls_main.append(experiment.params_to_cif())

            f_data = os.path.join(f_dir, f"{experiment.data_name:}_data.rcif")
            ls_data = []
            ls_data.append("\ndata_{:}".format(experiment.data_name))
            ls_data.append(experiment.data_to_cif())
            with open(f_data, 'w') as fid:
                fid.write("\n".join(ls_data))
            f_calc = os.path.join(f_dir, f"{experiment.data_name:}_calc.rcif")
            ls_calc = []
            ls_calc.append("\ndata_{:}".format(experiment.data_name))
            ls_calc.append(experiment.calc_to_cif())
            with open(f_calc, 'w') as fid:
                fid.write("\n".join(ls_calc))
        with open(f_main, 'w') as fid:
            fid.write("\n".join(ls_main))

    def estimate_f_mag_for_diffrn(
            self, maxAbsFM: float = 5, deltaFM: float = 0.001,
            flag_lambdaover2_correction: bool = False):
        """Estimate magnetic structure factor for diffrn experiments."""
        l_res = []
        experiments = self.experiments()
        for experiment in experiments:
            if isinstance(experiment, Diffrn):
                crystal = getattr(self, f"crystal_{experiment.phase.label:}")
                res = experiment.estimate_FM(
                    crystal, maxAbsFM=maxAbsFM, deltaFM=deltaFM,
                    flag_lambdaover2_correction=bool(
                        flag_lambdaover2_correction))
                l_res.append(res)
        return l_res
# s_cont = """
# global_
#  data_Fe3O4
#  _cell_angle_alpha 90.0
#  _cell_angle_beta 90.0
#  _cell_angle_gamma 90.0
#  _cell_length_a 8.56212()
#  _cell_length_b 8.56212
#  _cell_length_c 8.56212
#  _space_group_it_coordinate_system_code 2
#  _space_group_IT_number    227
#  loop_
#  _atom_site_adp_type
#  _atom_site_B_iso_or_equiv
#  _atom_site_fract_x
#  _atom_site_fract_y
#  _atom_site_fract_z
#  _atom_site_label
#  _atom_site_occupancy
#  _atom_site_type_symbol
#   Uani 0.0 0.125 0.125 0.125 Fe3A 1.0 Fe3+
#   Uani 0.0 0.5 0.5 0.5 Fe3B 1.0 Fe3+
#   Uiso 0.0 0.25521 0.25521 0.25521 O1 1.0 O2-
#  loop_
#  _atom_type_scat_length_neutron
#  _atom_type_symbol
#    0.945 Fe3+
#   0.5803 O2-
#  loop_
#  _atom_site_aniso_U_11
#  _atom_site_aniso_U_12
#  _atom_site_aniso_U_13
#  _atom_site_aniso_U_22
#  _atom_site_aniso_U_23
#  _atom_site_aniso_U_33
#  _atom_site_aniso_label
#   0.0 0.0 0.0 0.0 0.0 0.0 Fe3A
#   0.0 0.0 0.0 0.0 0.0 0.0 Fe3B
#  loop_
#  _atom_site_scat_label
#  _atom_site_scat_lande
#  Fe3A 2.0
#  Fe3B 2.0
#  loop_
#  _atom_site_susceptibility_label
#  _atom_site_susceptibility_chi_type
#  _atom_site_susceptibility_chi_11
#  _atom_site_susceptibility_chi_12
#  _atom_site_susceptibility_chi_13
#  _atom_site_susceptibility_chi_22
#  _atom_site_susceptibility_chi_23
#  _atom_site_susceptibility_chi_33
#   Fe3A Cani -3.468(74) 0.0 0.0 -3.468 0.0 -3.468
#   Fe3B Cani 3.041      0.0 0.0  3.041 0.0  3.041
#  data_mono
#  _setup_wavelength     0.840
#  _setup_field          1.000
#  _diffrn_radiation_polarization 1.0
#  _diffrn_radiation_efficiency   1.0
#  _extinction_mosaicity 100.0
#  _extinction_radius    50.0
#  _extinction_model     gauss
#  _diffrn_orient_matrix_UB_11 6.59783
#  _diffrn_orient_matrix_UB_12 -6.99807
#  _diffrn_orient_matrix_UB_13 3.3663
#  _diffrn_orient_matrix_UB_21 2.18396
#  _diffrn_orient_matrix_UB_22 -2.60871
#  _diffrn_orient_matrix_UB_23 -9.5302
#  _diffrn_orient_matrix_UB_31 7.4657
#  _diffrn_orient_matrix_UB_32 6.94702
#  _diffrn_orient_matrix_UB_33 -0.18685
#  _phase_label  Fe3O4
#  loop_
#  _diffrn_refln_index_h
#  _diffrn_refln_index_k
#  _diffrn_refln_index_l
#  _diffrn_refln_fr
#  _diffrn_refln_fr_sigma
#  0 0 8 0.64545 0.01329
#  2 0 6 1.75682 0.04540
#  0 2 6 1.67974 0.03711
# """
# obj = RhoChi.from_cif(s_cont)
# print(obj.diffrn_mono.diffrn_orient_matrix.cell)
# for var_name in obj.get_variable_names():
#     print(var_name)
#     obj.set_variable_by_name(var_name, 4)
# obj.form_object()
# print(obj.crystal_fe3o4)
